use core::{
	fmt::{self},
	ops::{Add, AddAssign, Sub, SubAssign},
};

/// A canonical 32-bit virtual memory address.
#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[repr(transparent)]
pub struct VirtAddr(u32);

impl VirtAddr {
	/// Create a new [`VirtAddr`] from a [`u32`].
	#[inline]
	pub const fn new(addr: u32) -> Self {
		Self(addr)
	}

	/// Converts this address to a `u32`.
	#[inline]
	pub const fn as_u32(self) -> u32 {
		self.0
	}
}

impl fmt::Debug for VirtAddr {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_tuple("VirtAddr")
			.field(&format_args!("{:#x}", self.0))
			.finish()
	}
}

impl fmt::Binary for VirtAddr {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		fmt::Binary::fmt(&self.0, f)
	}
}

impl fmt::LowerHex for VirtAddr {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		fmt::LowerHex::fmt(&self.0, f)
	}
}

impl fmt::Octal for VirtAddr {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		fmt::Octal::fmt(&self.0, f)
	}
}

impl fmt::UpperHex for VirtAddr {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		fmt::UpperHex::fmt(&self.0, f)
	}
}

impl fmt::Pointer for VirtAddr {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		fmt::Pointer::fmt(&(self.0 as *const ()), f)
	}
}

impl Add<u32> for VirtAddr {
	type Output = Self;
	#[inline]
	fn add(self, rhs: u32) -> Self::Output {
		VirtAddr(self.0.checked_add(rhs).unwrap())
	}
}

impl AddAssign<u32> for VirtAddr {
	#[inline]
	fn add_assign(&mut self, rhs: u32) {
		*self = *self + rhs;
	}
}

impl Sub<u32> for VirtAddr {
	type Output = Self;
	#[inline]
	fn sub(self, rhs: u32) -> Self::Output {
		VirtAddr(self.0.checked_sub(rhs).unwrap())
	}
}

impl SubAssign<u32> for VirtAddr {
	#[inline]
	fn sub_assign(&mut self, rhs: u32) {
		*self = *self - rhs;
	}
}
