mod draw;
mod falling;

pub use falling::FallingTetromino;

use crate::drivers::vga::VgaColor;

#[repr(u8)]
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum TetrisCell {
	Empty,
	Filled { color: VgaColor },
}

pub struct TetrisBoard {
	board: [TetrisCell; TetrisBoard::BOARD_WIDTH * TetrisBoard::BOARD_HEIGHT],
}

impl TetrisBoard {
	const BOARD_WIDTH: usize = 10;
	const BOARD_HEIGHT: usize = 20;

	const CELL_SIZE: usize = 9;

	pub const fn new() -> Self {
		Self {
			board: [TetrisCell::Empty; TetrisBoard::BOARD_WIDTH * TetrisBoard::BOARD_HEIGHT],
		}
	}

	pub fn place_tetromino(&mut self, tetromino: FallingTetromino) {
		for (x, y) in tetromino.tiles() {
			let cell = self.get_cell_mut(x, y);
			if let Some(cell) = cell {
				*cell = TetrisCell::Filled {
					color: tetromino.color,
				};
			}
		}
	}

	/// Returns `false` if the given tetromino intersects a filled cell
	/// or exits the board boundary
	pub fn tetromino_valid(&self, tetromino: &FallingTetromino) -> bool {
		for (x, y) in tetromino.tiles() {
			let cell = self.get_cell(x, y);
			if cell != Some(&TetrisCell::Empty) {
				return false;
			}
		}

		return true;
	}

	pub fn get_cell(&self, x: usize, y: usize) -> Option<&TetrisCell> {
		if y >= TetrisBoard::BOARD_HEIGHT {
			return None;
		}

		if x >= TetrisBoard::BOARD_WIDTH {
			return None;
		}

		return Some(&self.board[y * TetrisBoard::BOARD_WIDTH + x]);
	}

	pub fn get_cell_mut(&mut self, x: usize, y: usize) -> Option<&mut TetrisCell> {
		if y >= TetrisBoard::BOARD_HEIGHT {
			return None;
		}

		if x >= TetrisBoard::BOARD_WIDTH {
			return None;
		}

		return Some(&mut self.board[y * TetrisBoard::BOARD_WIDTH + x]);
	}
}
